#include "libckpt.h"
#include "copyright.h"

#define	BUF_SIZE	8192

typedef struct listtype {
  char *start;
  unsigned long size;
  char *stop;
  int fnum;
  int offset;
  struct listtype *next;
} *ListType;

#ifdef __STDC__

void _ckpt_coalesce_files(char *);
static int coa_open_files(char *, int []);
static ListType coa_read_files(int, int[]);
static ListType coa_insert(ListType, char *, unsigned long, int, off_t);
static ListType coa_mknode(char *, unsigned long, int, off_t);
static void coa_insert_node(ListType, ListType, ListType, char *, unsigned long, int, off_t);
static void coa_combine_files(char *, int, int [], ListType);
static void coa_transfer_chunk(ListType, int [], int);

#else

void _ckpt_coalesce_files();
static int coa_open_files();
static ListType coa_read_files();
static ListType coa_insert();
static ListType coa_mknode();
static void coa_insert_node();
static void coa_combine_files();
static void coa_transfer_chunk();

#endif /* __STDC__ */

void _ckpt_coalesce_files(dir)
char *dir;
{

  int fd[_CKPT_FTABLE_SIZE]; /* array of file descriptors */
  int fnum;                  /* number of checkpoint files */
  ListType list;

  if ( _ckpt_verbose_flag )
    fprintf ( stderr, "CKP %d : coalescing files\n", time(0) );

  fnum = coa_open_files(dir, fd);

  list = coa_read_files(fnum, fd);

  coa_combine_files(dir, fnum, fd, list);

  if ( _ckpt_verbose_flag )
    fprintf ( stderr, "CKP %d : done coalescing files\n", time(0) );

}

static int coa_open_files(dir, fd)
char *dir;
int fd[];
{

  int fnum = 0;
  char fname[MAXLEN];

  if ( strlen(dir) > MAXLEN - 10 ) {
    fprintf(stderr, "pathname length too long\n");
    return -1;
  }

  sprintf ( fname, "%s/%s.%d", dir, _ckpt_filename, fnum );

  while ( (fd[fnum] = OPEN(fname, O_RDONLY)) != -1 ) {
    fprintf(stderr, "opened %s\n", fname);
    fnum++;
    sprintf ( fname, "%s/%s.%d", dir, _ckpt_filename, fnum );
  }

  return fnum;

}

static ListType coa_read_files(fnum, fd)
int fnum;
int fd[];
{

  ListType list;
  int first = 0;
  unsigned long nchunks;
  int i;
  char *addr;
  unsigned long size;

  list = NULL;

  for (fnum--; fnum >= 0; fnum--) {

    lseek(fd[fnum], 2 * POINTER_SIZE, SEEK_SET);

    read(fd[fnum], (char *)&nchunks, POINTER_SIZE);
    fprintf(stderr, "reading file #%d, chunks = %d\n", fnum, nchunks);
    for ( i = 0; i < nchunks - first; i++ ) {
      read ( fd[fnum], (char *)&addr, POINTER_SIZE );
      read ( fd[fnum], (char *)&size, POINTER_SIZE );
      list = coa_insert(list, addr, size, fnum, lseek(fd[fnum], 0, SEEK_CUR));
      lseek ( fd[fnum], size, SEEK_CUR );
    }

/*
 *  On Sparc's, we have to skip over the file area as well as the stack.
 *  On other architectures, we only skip over the stack.
 */

#ifdef sparc
    first = 2;
#else
    first = 1;
#endif

  }

  return list;

}

static ListType coa_insert(list, addr, size, fnum, offset)
ListType list;
char *addr;
unsigned long size;
int fnum;
off_t offset;
{

  ListType p1, p2;

  if ( list == NULL )
    list = coa_mknode(addr, size, fnum, offset);
  else {
    p1 = NULL;
    p2 = list;
    while (p1 || p2) {
      coa_insert_node(list, p1, p2, addr, size, fnum, offset);
      if (p1 && p2) {
        p1 = p1->next;
        p2 = p2->next;
      }
      else if (p1 && !p2) {
        p1 = p1->next;
      }
      else if (!p1 && p2) {
        p1 = list;
        p2 = p2->next;
      }
    }
  }

  return list;

}

static ListType coa_mknode(addr, size, fnum, offset)
char *addr;
unsigned long size;
int fnum;
off_t offset;
{

  ListType ptr;

  ptr = (ListType)malloc(sizeof(struct listtype));

  ptr->start = addr;
  ptr->size = size;
  ptr->stop = addr + size;
  ptr->fnum = fnum;
  ptr->offset = offset;
  ptr->next = NULL;

  return ptr;

}

static void coa_insert_node(list, p1, p2, addr, size, fnum, offset)
ListType list, p1, p2;
char *addr;
unsigned long size;
int fnum;
off_t offset;
{

  ListType temp;
  char *new_addr;
  unsigned long new_size;
  off_t new_offset;

  if ( ! p1 ) {
    if ( addr < p2->start ) {
      if ( addr+size < addr ) {
        temp = coa_mknode(addr, size, fnum, offset);
        temp->next = p2;
        list = temp;
      }
      else {
        new_size = p2->start - addr;
        temp = coa_mknode(addr, new_size, fnum, offset);
        temp->next = p2;
        list = temp;
      }
    }
  }

  else if ( ! p2 ) {
    if ( addr < p1->start + size ) {
      if ( addr + size > p1->start + size ) {
        new_addr = p1->start + p1->size;
        new_size = addr + size - new_addr;
        new_offset = offset + new_addr - addr;
        temp = coa_mknode(new_addr, new_size, fnum, new_offset);
        p1->next = temp;
      }
    }
    else {
      temp = coa_mknode(addr, size, fnum, offset);
      p1->next = temp;
    }
  }

  else { /* p1 and p2 are non-NULL */
    if ( (addr + size > p1->start + p1->size) && (addr < p2->start) ) {
      if ( addr < p1->start + p1->size ) {
        new_addr = p1->start + p1->size;
        new_offset = offset + new_addr - addr;
      }
      else {
        new_addr = addr;
        new_offset = offset;
      }
      if ( addr + size > p2->start )
        new_size = p2->start - addr;
      else
        new_size = size;
      temp = coa_mknode(new_addr, new_size, fnum, new_offset);
      temp->next = p2;
      p1->next = temp;
    }
  }

}

static void coa_combine_files(dir, fnum, fd, list)
char *dir;
int fnum;
int fd[];
ListType list;
{

  int new_fd;
  char fname[MAXLEN];
  char temp_name[MAXLEN];
  char new_name[MAXLEN];
  char *start;
  char *stop;
  unsigned long nchunks = 0;

/*
 *  open a temp file, dir/ckpt.temp
 *  ckpt is the default base filename or the name specified by the
 *  user's call to ckpt_name(), and dir is the directory where the
 *  checkpoint files are located.
 */

  sprintf(temp_name, "%s/%s.temp", dir, _ckpt_filename);
  remove(temp_name);
  new_fd = OPEN_M(temp_name, O_WRONLY | O_CREAT, 0644);

  lseek(fd[fnum-1], 0, SEEK_SET);
  read(fd[fnum-1], (char *)&start, POINTER_SIZE);
  read(fd[fnum-1], (char *)&stop, POINTER_SIZE);

/*
 *  Write the header info to the temp file
 */

  write(new_fd, (char *)&start, POINTER_SIZE);
  write(new_fd, (char *)&stop, POINTER_SIZE);
  write(new_fd, (char *)&nchunks, POINTER_SIZE);

/*
 *  Write all the needed chunks from the various checkpoint files
 *  to the temp file.
 */

  while ( list ) {
    coa_transfer_chunk(list, fd, new_fd);
    list = list->next;
    nchunks++;
  }

/*
 *  Write the number of chunks to the temp file.
 */

  lseek(new_fd, 2 * POINTER_SIZE, SEEK_SET);
  write(new_fd, (char *)&nchunks, POINTER_SIZE);

/*
 *  Remove the old checkpoint files.
 */

  for ( fnum--; fnum >= 0; fnum-- ) {
    sprintf ( fname, "%s%s%d", dir, "/ckpt.", fnum );
    remove ( fname );
  }

/*
 *  Rename the tempfile, ckpt.temp, to ckpt.0
 */

  sprintf ( new_name, "%s/%s.0", _ckpt_dir, _ckpt_filename );
  rename ( temp_name, new_name );

}

static void coa_transfer_chunk(list, fd, new_fd)
ListType list;
int fd[];
int new_fd;
{

  char buf[BUF_SIZE];
  unsigned long size;

  size = list->size;

  lseek(fd[list->fnum], list->offset, SEEK_SET);
  write(new_fd, (char *)&(list->start), POINTER_SIZE);
  write(new_fd, (char *)&(list->size), POINTER_SIZE);

  while ( size >= BUF_SIZE ) {
    read(fd[list->fnum], (char *)buf, BUF_SIZE);
    write(new_fd, (char *)buf, BUF_SIZE);
    size -= BUF_SIZE;
  }

  if ( size > 0 ) {
    read(fd[list->fnum], (char *)buf, size);
    write(new_fd, (char *)buf, size);
  }

}
