#include "copyright.h"

/* 
 * $Source: /n/fs/vd/jsp/src/jgraph/RCS/list.c,v $
 * $Revision: 5.4 $
 * $Date: 91/06/06 16:14:15 $
 * $Author: jsp $
 */

#include <stdio.h>    /* Basic includes and definitions */
#include <malloc.h>
#include "dlist.h"

#define boolean int
#define TRUE 1
#define FALSE 0

#ifdef __alpha
void *malloc();
#endif

/*---------------------------------------------------------------------*
 * PROCEDURES FOR MANIPULATING DOUBLY LINKED LISTS 
 * Each list contains a sentinal node, so that     
 * the first item in list l is l->flink.  If l is  
 * empty, then l->flink = l->blink = l.            
 *---------------------------------------------------------------------*/

Dlist make_dl()
{
  Dlist d;

  d = (Dlist) malloc (sizeof(struct dlist));
  d->flink = d;
  d->blink = d;
  return d;
}
 
dl_insert_b(node)	/* Inserts a new node before the given one */
Dlist node;
{
  Dlist last_node, new;

  new = (Dlist) malloc (sizeof(struct dlist));

  last_node = node->blink;

  node->blink = new;
  last_node->flink = new;
  new->blink = last_node;
  new->flink = node;
}

dl_insert_list_b(node, list_to_insert)
Dlist node;
Dlist list_to_insert;
{
  Dlist last_node, f, l;

  if (dl_empty(list_to_insert)) {
    free(list_to_insert);
    return;
  }
  f = list_to_insert->flink;
  l = list_to_insert->blink;
  last_node = node->blink;

  node->blink = l;
  last_node->flink = f;
  f->blink = last_node;
  l->flink = node;
  free(list_to_insert);
}

dl_delete_node(item)		/* Deletes an arbitrary iterm */
Dlist item;
{
  item->flink->blink = item->blink;
  item->blink->flink = item->flink;
  free(item);
}

dl_delete_list(l)
Dlist l;
{
  Dlist d, next_node;

  d = l->flink;
  while(d != l) {
    next_node = d->flink;
    free(d);
    d = next_node;
  }
  free(d);
}

Mlist make_ml()
{
  Mlist m;
  Dlist d;

  m = (Mlist) malloc(sizeof(struct mlist));
  m->main = make_dl();
  dl_insert_b(m->main);
  dl_insert_b(m->main);
  d = m->main->flink;
  d->addr = (char *) 0;
  d->stop = (char *) 0;
  d->size = 0;
  d = d->flink;
#ifdef __alpha
  d->addr = (char *) 0xffffffffffffffff;
  d->stop = (char *) 0xffffffffffffffff;
#else
  d->addr = (char *) 0xffffffff;
  d->stop = (char *) 0xffffffff;
#endif
  d->size = 0;

  m->ptr = m->main->flink;
  return m;
}

ml_free_list(m)
Mlist m;
{
  Dlist d;
  d = m->main->flink;
  while (d != m->main) {
    d = d->flink;
    free(d->blink);
  }
  free(m->main);
  free(m);
}

ml_add(m, d, addr, size)
Mlist m;
Dlist d;
char *addr;
unsigned long size;
{
  d = d->flink;
  dl_insert_b(d);
  d = d->blink;
  d->addr = addr;
  d->size = size;
  d->stop = d->addr + d->size;
  m->ptr = d;
}

ml_delete(m, d)
Mlist m;
Dlist d;
{
  if (m->ptr == d) m->ptr = d->blink;
  dl_delete_node(d);
}

Dlist ml_find(m, addr)
Mlist m;
char *addr;
{
  Dlist d;
  
  d = m->ptr;
  if (d->addr == addr) return d;
  while (d->addr <= addr) {
    d = d->flink;
    if (d->addr > addr) {
      m->ptr = d->blink;
      return m->ptr;
    }
  }
  while (d->addr > addr) {
    d = d->blink;
    if (d->addr <= addr) {
      m->ptr = d;
      return d;
    }
  }
}
