#include "libckpt.h"
#include "copyright.h"

static long _ckpt_seekptr;

void ckpt_recover_(s)
char *s;
{

  pname_type path;
  char fname[MAXLEN];
  int fd;

  _ckpt_filename = strdup(s);

  path = _ckpt_get_path();
  sprintf ( fname, "%s/%s.0", path.name, _ckpt_filename );

  fd = OPEN(fname, O_RDONLY);

  if ( fd == -1 ) {
    fprintf(stderr, "=recover flag specified, but no checkpoint ");
    fprintf(stderr, "files exist in directory %s\n", _ckpt_dir);
    _ckpt_flag = FALSE;
    exit(-1);
  }

  CLOSE(fd);

  _ckpt_recover(path);

}

void _ckpt_recover(path)
pname_type path;
{

  int fn;

#ifdef _pvm
  pvm_exit();
#endif

  fn = recover_heap(path);

  recover_stack(fn,path);

}

static int recover_heap(path)
pname_type path;
{

  int fd;
  char *addr, *top_stack, *stop_addr;
  long size, nchunks;
  int i;
  char filename[MAXLEN];
  int fn = 0;

  while (1) {

    sprintf(filename, "%s/%s.%d", path.name, _ckpt_filename, fn);
    fn++;

    fd = OPEN(filename, O_RDONLY);

    if (fd == -1) break;

    read(fd, (char *)&top_stack, POINTER_SIZE);
    read(fd, (char *)&stop_addr, POINTER_SIZE);
    read(fd, (char *)&nchunks, POINTER_SIZE);

    brk(stop_addr);

    for (i = 0; i < nchunks - 1; i++) {
      read(fd, (char *)&addr, POINTER_SIZE);
      read(fd, (char *)&size, POINTER_SIZE);
      read(fd, (char *)addr, size);
    }

    _ckpt_seekptr = lseek(fd, 0, SEEK_CUR);

    CLOSE(fd);

  }

  return fn - 2;

}

static void recover_stack(fn, path)
int fn;
pname_type path;
{

  caddr_t top_stack;
  int fd;
  char filename[MAXLEN];

  sprintf(filename, "%s/%s.%d", path.name, _ckpt_filename, fn);
  fd = OPEN(filename ,O_RDONLY);

  read(fd, (char *)&top_stack, POINTER_SIZE);

  CLOSE(fd);

  restore_memory(top_stack, fn, path);

}

static void restore_memory(top_stack, fn, path)
char *top_stack;
int fn;
pname_type path;
{

  int idummy;
  char *new_top_stack;
  char *addr;
  long size;
  int fd;
  char filename[MAXLEN];

  new_top_stack = (char *)&idummy;

#ifdef STACK_GROWS_UP
  if (new_top_stack >= top_stack - OFFSET)
#else
  if (new_top_stack <= top_stack - OFFSET)
#endif
    restore_memory( top_stack, fn, path );

  sprintf(filename, "%s/%s.%d", path.name, _ckpt_filename, fn);
  fd = OPEN(filename ,O_RDONLY);

  lseek(fd, _ckpt_seekptr, SEEK_SET);

  read(fd, (char *)&addr, POINTER_SIZE);
  read(fd, (char *)&size, POINTER_SIZE);
  read(fd, (char *)addr, size);

  CLOSE(fd);

  longjmp(_ckpt_env,1);

}
