#define SYSTEM_C
#include "libckpt.h"
#include "copyright.h"

int
open(filename, attribute, mode)
char *filename;
int attribute;
int mode;
{
  char *strptr;
  int fd;

  do
    fd = OPEN_M(filename, attribute, mode);
  while (fd == -1 && errno == EINTR);

  if ( _ckpt_flag && fd != -1 ) {

    file_insert(filename, attribute, mode, fd, 0);
  }

  return fd;
}

int
dup(fd)
int fd;
{
  int new_fd;

  do
    new_fd = DUP(fd);
  while (fd == -1 && errno == EINTR);

  if ( _ckpt_flag && fd != -1 & _ckpt_ftable[fd].inuse)

    file_insert(_ckpt_ftable[fd].filename,
		_ckpt_ftable[fd].attribute,
		_ckpt_ftable[fd].mode, new_fd, new_fd);

  return fd;
}

int
dup2(fd1, fd2)
int fd1, fd2;
{
  int new_fd;
  ftable_entry ptr;

  if (_ckpt_ftable[fd2].inuse) file_delete(fd2);
  
  do
    new_fd = DUP2(fd1, fd2);
  while (new_fd == -1 && errno == EINTR);


  if ( _ckpt_flag && new_fd != -1  && _ckpt_ftable[fd1].inuse)

    file_insert(_ckpt_ftable[fd1].filename,
		_ckpt_ftable[fd1].attribute,
		_ckpt_ftable[fd1].mode, new_fd, new_fd);

  return new_fd;
}

int
close(fd)
int fd;
{
  ftable_entry *fptr, *free_ptr;
  int result;

  do
    result = CLOSE(fd);
  while (result == -1 && errno == EINTR);

  if ( _ckpt_flag && result == 0 && _ckpt_ftable[fd].inuse)

    file_delete(fd);

  return result;

}

static void
file_insert(filename, attribute, mode, fd, dup_of)
char *filename;
int attribute, mode, fd, dup_of;
{
    char *strptr;
    ftable_entry *new_ptr, *ptr;

    strptr = (char *)malloc(strlen(filename) + 1);
    strcpy(strptr,filename);

    new_ptr = &_ckpt_ftable[fd];

    new_ptr -> filename  	= strptr;
    new_ptr -> attribute 	= attribute;
    new_ptr -> mode 		= mode;
    new_ptr -> fd 		= fd;
    new_ptr -> next 		= NULL;
    new_ptr -> inuse 		= 1;

    if (dup_of > 0) {
        new_ptr->dup_list 		= _ckpt_ftable[dup_of].dup_list;
        _ckpt_ftable[dup_of].dup_list 	= new_ptr;
    }  else
	new_ptr -> dup_list 		= new_ptr;
}

static void file_delete(fd)
int fd;
{
    ftable_entry *fptr, *free_ptr;

    fptr  = &_ckpt_ftable[fd]; 
    fptr -> dup_list = fptr -> dup_list -> dup_list; 
    free(fptr->filename); 
    fptr -> inuse = 0; 
}

int
read(fd, buf, nbyte)
int fd;
char *buf;
int nbyte;
{

  int rtime;
  char *ptr;
  char *stop;
  int result;

/*
 *  The following had to be added because if the break was exactly on
 *  a page boundary, and this last page was written during checkpointing,
 *  then during recover the while loop tried to touch a page that was
 *  read only, and a SEGV occured.
 */

#if !defined(__alpha) && !defined(__linux)

  stop = (char *) min ( (int)sbrk(0), (int)buf + nbyte + 1 );

  ptr = buf;
  *ptr = *ptr;

/*
 *  The following works because (ptr % GPAGESIZE) is 0 if ptr is
 *  on a page boundary, so ptr goes up to the next page boundary.
 */

  ptr += GPAGESIZE - ((long)ptr % GPAGESIZE);

  while ( ptr < stop ) {
    *ptr = *ptr;
    ptr += GPAGESIZE;
  }

#endif /* ! __alpha && ! __linux */

  do
    result = READ(fd, buf, nbyte);
  while (result == -1 && errno == EINTR);

  return result;

}

int
write(fd, buf, nbyte)
int fd;
char *buf;
int nbyte;
{

  int rtime;
  int result;

  do
    result = WRITE(fd, buf, nbyte);
  while (result == -1 && errno == EINTR);

  return result;

}
